@@ -7,10 +7,14 @@ use std::{
 };
 
 use ipnetwork::IpNetwork;
+use log::{error, info, warn};
 use tokio::sync::RwLockReadGuard;
 use trust_dns_resolver::{
     config::{NameServerConfigGroup, ResolverOpts},
-    proto::rr::{dnssec::SupportedAlgorithms, RecordSet},
+    proto::{
+        error::ProtoError,
+        rr::{dnssec::SupportedAlgorithms, rdata::SOA, RecordSet},
+    },
     IntoName,
 };
 
@@ -20,28 +24,68 @@ use trust_dns_server::{
     store::forwarder::ForwardAuthority,
     store::{forwarder::ForwardConfig, in_memory::InMemoryAuthority},
 };
-use zerotier_central_api::{apis::configuration::Configuration, models::Member};
+use zerotier_central_api::{
+    apis::configuration::Configuration,
+    models::{Member, Network},
+};
 
 use crate::{
+    addresses::Calculator,
     hosts::{parse_hosts, HostsFile},
     utils::{parse_member_name, ToHostname},
 };
 
+pub(crate) trait ToPointerSOA {
+    fn to_ptr_soa_name(self) -> Result<Name, ProtoError>;
+}
+
+impl ToPointerSOA for IpNetwork {
+    fn to_ptr_soa_name(self) -> Result<Name, ProtoError> {
+        // how many bits in each ptr octet
+        let octet_factor = match self {
+            IpNetwork::V4(_) => 8,
+            IpNetwork::V6(_) => 4,
+        };
+
+        Ok(self
+            .network()
+            .into_name()?
+            // round off the subnet, account for in-addr.arpa.
+            .trim_to((self.prefix() as usize / octet_factor) + 2))
+    }
+}
+
+pub(crate) trait ToWildcard {
+    fn to_wildcard(self, count: u8) -> Name;
+}
+
+impl ToWildcard for Name {
+    fn to_wildcard(self, count: u8) -> Name {
+        let mut name = Self::from_str("*").unwrap();
+        for _ in 0..count {
+            name = name.append_domain(&Self::from_str("*").unwrap());
+        }
+
+        name.append_domain(&self).into_wildcard()
+    }
+}
+
 pub(crate) type TokioZTAuthority = Arc<tokio::sync::RwLock<ZTAuthority>>;
 pub(crate) type Authority = Box<Arc<RwLock<InMemoryAuthority>>>;
-pub(crate) type PtrAuthority = Option<Authority>;
+pub(crate) type PtrAuthorityMap = HashMap<IpNetwork, Authority>;
 
-pub(crate) fn new_ptr_authority(ip: IpNetwork) -> Result<PtrAuthority, anyhow::Error> {
-    Ok(match ip {
-        IpNetwork::V4(ip) => Some(Box::new(Arc::new(RwLock::new(InMemoryAuthority::empty(
-            ip.network()
-                .into_name()?
-                .trim_to((ip.prefix() as usize / 8) + 2),
-            trust_dns_server::authority::ZoneType::Primary,
-            false,
-        ))))),
-        IpNetwork::V6(_) => None,
-    })
+pub(crate) fn new_ptr_authority(ip: IpNetwork) -> Result<Authority, anyhow::Error> {
+    let domain_name = ip.to_ptr_soa_name()?;
+
+    let mut authority = InMemoryAuthority::empty(
+        domain_name.clone(),
+        trust_dns_server::authority::ZoneType::Primary,
+        false,
+    );
+
+    set_soa(&mut authority, domain_name);
+
+    Ok(Box::new(Arc::new(RwLock::new(authority))))
 }
 
 pub(crate) async fn find_members(zt: TokioZTAuthority) {
@@ -58,27 +102,33 @@ pub(crate) async fn find_members(zt: TokioZTAuthority) {
             .expect("Could not configure authority from hosts file");
 
         match get_members(zt.read().await).await {
-            Ok(members) => match zt.write().await.configure_members(members) {
+            Ok((network, members)) => match zt.write().await.configure_members(network, members) {
                 Ok(_) => {}
                 Err(e) => {
-                    eprintln!("error configuring authority: {}", e)
+                    error!("error configuring authority: {}", e)
                 }
             },
             Err(e) => {
-                eprintln!("error syncing members: {}", e)
+                error!("error syncing members: {}", e)
             }
         }
     }
 }
 
-async fn get_members(zt: RwLockReadGuard<'_, ZTAuthority>) -> Result<Vec<Member>, anyhow::Error> {
+async fn get_members(
+    zt: RwLockReadGuard<'_, ZTAuthority>,
+) -> Result<(Network, Vec<Member>), anyhow::Error> {
     let config = zt.config.clone();
     let network = zt.network.clone();
 
-    Ok(
+    let members =
         zerotier_central_api::apis::network_member_api::get_network_member_list(&config, &network)
-            .await?,
-    )
+            .await?;
+
+    let network =
+        zerotier_central_api::apis::network_api::get_network_by_id(&config, &network).await?;
+
+    Ok((network, members))
 }
 
 fn prune_records(authority: Authority, written: Vec<Name>) -> Result<(), anyhow::Error> {
@@ -89,15 +139,15 @@ fn prune_records(authority: Authority, written: Vec<Name>) -> Result<(), anyhow:
 
     let rr = lock.records_mut();
 
-    for (rrkey, _) in rr.clone() {
+    for (rrkey, rs) in rr.clone() {
         let key = &rrkey.name().into_name()?;
-        if !written.contains(key) {
+        if !written.contains(key) && rs.record_type() != RecordType::SOA {
             rrkey_list.push(rrkey);
         }
     }
 
     for rrkey in rrkey_list {
-        eprintln!("Removing expired record {}", rrkey.name());
+        warn!("Removing expired record {}", rrkey.name());
         rr.remove(&rrkey);
     }
 
@@ -144,7 +194,7 @@ fn upsert_address(
         {
             let mut address = Record::with(fqdn.clone(), rt, 60);
             address.set_rdata(rdata.clone());
-            eprintln!("Adding new record {}: ({})", fqdn.clone(), rdata.clone());
+            info!("Adding new record {}: ({})", fqdn.clone(), rdata.clone());
             authority.upsert(address, serial);
         }
     }
@@ -155,7 +205,7 @@ fn set_ptr_record(
     ip_name: Name,
     canonical_name: Name,
 ) {
-    eprintln!(
+    warn!(
         "Replacing PTR record {}: ({})",
         ip_name.clone(),
         canonical_name
@@ -200,7 +250,7 @@ fn set_ip_record(
 
 fn configure_ptr(
     authority: Authority,
-    ip: IpAddr,
+    ip: Name,
     canonical_name: Name,
 ) -> Result<(), anyhow::Error> {
     let lock = &mut authority
@@ -209,40 +259,57 @@ fn configure_ptr(
 
     match lock
         .records()
-        .get(&RrKey::new(ip.into_name()?.into(), RecordType::PTR))
+        .get(&RrKey::new(ip.clone().into(), RecordType::PTR))
     {
         Some(records) => {
             if !records
                 .records(false, SupportedAlgorithms::all())
                 .any(|rec| rec.rdata().eq(&RData::PTR(canonical_name.clone())))
             {
-                set_ptr_record(lock, ip.into_name()?, canonical_name.clone());
+                set_ptr_record(lock, ip.clone(), canonical_name.clone());
             }
         }
-        None => set_ptr_record(lock, ip.into_name()?, canonical_name.clone()),
+        None => set_ptr_record(lock, ip.clone(), canonical_name.clone()),
     }
     Ok(())
 }
 
+pub(crate) fn set_soa(authority: &mut InMemoryAuthority, domain_name: Name) {
+    let mut soa = Record::new();
+    soa.set_name(domain_name.clone());
+    soa.set_rr_type(RecordType::SOA);
+    soa.set_rdata(RData::SOA(SOA::new(
+        domain_name.clone(),
+        Name::from_str("administrator")
+            .unwrap()
+            .append_domain(&domain_name.clone()),
+        authority.serial() + 1,
+        30,
+        0,
+        -1,
+        0,
+    )));
+    authority.upsert(soa, authority.serial() + 1);
+}
+
 pub(crate) fn init_trust_dns_authority(domain_name: Name) -> Authority {
-    Box::new(Arc::new(RwLock::new(InMemoryAuthority::empty(
+    let mut authority = InMemoryAuthority::empty(
         domain_name.clone(),
         trust_dns_server::authority::ZoneType::Primary,
         false,
-    ))))
+    );
+
+    set_soa(&mut authority, domain_name.clone());
+    Box::new(Arc::new(RwLock::new(authority)))
 }
 
-pub(crate) async fn init_catalog(zt: TokioZTAuthority) -> Result<Catalog, std::io::Error> {
+pub(crate) async fn init_catalog(zt: TokioZTAuthority) -> Result<Catalog, anyhow::Error> {
     let read = zt.read().await;
 
     let mut catalog = Catalog::default();
     catalog.upsert(read.domain_name.clone().into(), read.authority.box_clone());
-
-    if read.ptr_authority.is_some() {
-        let unwrapped = read.ptr_authority.clone().unwrap();
-        catalog.upsert(unwrapped.origin(), unwrapped.box_clone());
-    } else {
-        println!("PTR records are not supported on IPv6 networks (yet!)");
+    for (network, authority) in read.ptr_authority_map.clone() {
+        catalog.upsert(network.to_ptr_soa_name()?.into(), authority.box_clone());
     }
 
     drop(read);
@@ -276,9 +343,149 @@ pub(crate) async fn init_catalog(zt: TokioZTAuthority) -> Result<Catalog, std::i
     Ok(catalog)
 }
 
+pub(crate) struct ZTRecord {
+    fqdn: Name,
+    canonical_name: Option<Name>,
+    ptr_name: Name,
+    ips: Vec<IpAddr>,
+    wildcard_everything: bool,
+    member: Member,
+}
+
+impl ZTRecord {
+    pub(crate) fn new(
+        member: &Member,
+        sixplane: Option<IpNetwork>,
+        rfc4193: Option<IpNetwork>,
+        domain_name: Name,
+        wildcard_everything: bool,
+    ) -> Result<Self, anyhow::Error> {
+        let member_name = format!(
+            "zt-{}",
+            member
+                .clone()
+                .node_id
+                .expect("Node ID for member does not exist")
+        );
+
+        let fqdn = member_name.clone().to_fqdn(domain_name.clone())?;
+
+        // this is default the zt-<member id> but can switch to a named name if
+        // tweaked in central. see below.
+        let mut canonical_name = None;
+        let mut ptr_name = fqdn.clone();
+
+        if let Some(name) = parse_member_name(member.name.clone(), domain_name.clone()) {
+            canonical_name = Some(name.clone());
+            ptr_name = name;
+        }
+
+        let mut ips: Vec<IpAddr> = member
+            .clone()
+            .config
+            .expect("Member config does not exist")
+            .ip_assignments
+            .expect("IP assignments for member do not exist")
+            .into_iter()
+            .map(|s| IpAddr::from_str(&s).expect("Could not parse IP address"))
+            .collect();
+
+        if sixplane.is_some() {
+            ips.push(member.clone().sixplane()?.ip());
+        }
+
+        if rfc4193.is_some() {
+            ips.push(member.clone().rfc4193()?.ip());
+        }
+
+        Ok(Self {
+            member: member.clone(),
+            wildcard_everything,
+            fqdn,
+            canonical_name,
+            ptr_name,
+            ips,
+        })
+    }
+
+    pub(crate) fn insert_records(&self, records: &mut Vec<Name>) {
+        records.push(self.fqdn.clone());
+
+        for ip in self.ips.clone() {
+            records.push(ip.into_name().expect("Could not coerce IP into name"))
+        }
+
+        if self.canonical_name.is_some() {
+            records.push(self.canonical_name.clone().unwrap());
+            if self.wildcard_everything {
+                records.push(self.get_canonical_wildcard().unwrap())
+            }
+        }
+
+        if self.wildcard_everything {
+            records.push(self.fqdn.clone().to_wildcard(0))
+        }
+    }
+
+    pub(crate) fn get_canonical_wildcard(&self) -> Option<Name> {
+        if self.canonical_name.is_none() {
+            return None;
+        }
+
+        Some(self.canonical_name.clone().unwrap().to_wildcard(0))
+    }
+
+    pub(crate) fn insert_authority(&self, authority: &ZTAuthority) -> Result<(), anyhow::Error> {
+        authority.match_or_insert(self.fqdn.clone(), self.ips.clone());
+
+        if self.wildcard_everything {
+            authority.match_or_insert(self.fqdn.clone().to_wildcard(0), self.ips.clone());
+        }
+
+        if self.canonical_name.is_some() {
+            authority.match_or_insert(self.canonical_name.clone().unwrap(), self.ips.clone());
+            if self.wildcard_everything {
+                authority.match_or_insert(self.get_canonical_wildcard().unwrap(), self.ips.clone())
+            }
+        }
+
+        Ok(())
+    }
+
+    pub(crate) fn insert_member_ptr(
+        &self,
+        authority_map: PtrAuthorityMap,
+        _sixplane: Option<IpNetwork>,
+        rfc4193: Option<IpNetwork>,
+        records: &mut Vec<Name>,
+    ) -> Result<(), anyhow::Error> {
+        for ip in self.ips.clone() {
+            for (network, authority) in authority_map.clone() {
+                if network.contains(ip) {
+                    let ip = ip.into_name().expect("Could not coerce IP into name");
+                    records.push(ip.clone());
+                    configure_ptr(authority.clone(), ip, self.ptr_name.clone())?;
+                }
+            }
+        }
+
+        if let Some(rfc4193) = rfc4193 {
+            let ptr = self.member.clone().rfc4193()?.network().into_name()?;
+
+            records.push(ptr.clone());
+
+            if let Some(authority) = authority_map.get(&rfc4193) {
+                configure_ptr(authority.clone(), ptr, self.ptr_name.clone())?;
+            }
+        }
+
+        Ok(())
+    }
+}
+
 #[derive(Clone)]
 pub(crate) struct ZTAuthority {
-    ptr_authority: PtrAuthority,
+    ptr_authority_map: PtrAuthorityMap,
     authority: Authority,
     domain_name: Name,
     network: String,
@@ -286,6 +493,7 @@ pub(crate) struct ZTAuthority {
     hosts_file: Option<String>,
     hosts: Option<Box<HostsFile>>,
     update_interval: Duration,
+    wildcard_everything: bool,
 }
 
 impl ZTAuthority {
@@ -294,7 +502,7 @@ impl ZTAuthority {
         network: String,
         config: Configuration,
         hosts_file: Option<String>,
-        ptr_authority: PtrAuthority,
+        ptr_authority_map: PtrAuthorityMap,
         update_interval: Duration,
         authority: Authority,
     ) -> Self {
@@ -304,12 +512,17 @@ impl ZTAuthority {
             network,
             config,
             authority,
-            ptr_authority,
+            ptr_authority_map,
             hosts_file,
             hosts: None,
+            wildcard_everything: false,
         }
     }
 
+    pub(crate) fn wildcard_everything(&mut self) {
+        self.wildcard_everything = true;
+    }
+
     fn match_or_insert(&self, name: Name, newips: Vec<IpAddr>) {
         let rdatas: Vec<RData> = newips
             .clone()
@@ -330,6 +543,15 @@ impl ZTAuthority {
                 .get(&RrKey::new(name.clone().into(), rt))
                 .clone();
 
+            let ips = newips
+                .clone()
+                .into_iter()
+                .filter(|i| match i {
+                    IpAddr::V4(_) => rt == RecordType::A,
+                    IpAddr::V6(_) => rt == RecordType::AAAA,
+                })
+                .collect();
+
             match records {
                 Some(records) => {
                     let records = records.records(false, SupportedAlgorithms::all());
@@ -341,7 +563,7 @@ impl ZTAuthority {
                             &mut self.authority.write().expect("write lock"),
                             name.clone(),
                             rt,
-                            newips.clone(),
+                            ips,
                         );
                     }
                 }
@@ -351,67 +573,63 @@ impl ZTAuthority {
                         &mut self.authority.write().expect("write lock"),
                         name.clone(),
                         rt,
-                        newips.clone(),
+                        ips,
                     );
                 }
             }
         }
     }
 
-    fn configure_members(&mut self, members: Vec<Member>) -> Result<(), anyhow::Error> {
-        let mut records = Vec::new();
+    fn configure_members(
+        &mut self,
+        network: Network,
+        members: Vec<Member>,
+    ) -> Result<(), anyhow::Error> {
+        let mut records = vec![self.domain_name.clone()];
         if let Some(hosts) = self.hosts.to_owned() {
             self.prune_hosts();
             records.append(&mut hosts.values().flatten().map(|v| v.clone()).collect());
         }
 
+        let (mut sixplane, mut rfc4193) = (None, None);
+
+        let v6assign = network.config.clone().unwrap().v6_assign_mode;
+        if v6assign.is_some() {
+            let v6assign = v6assign.unwrap().clone();
+            if v6assign.var_6plane.unwrap_or(false) {
+                let s = network.clone().sixplane()?;
+                sixplane = Some(s);
+                records.push(s.to_ptr_soa_name()?);
+            }
+
+            if v6assign.rfc4193.unwrap_or(false) {
+                let s = network.clone().rfc4193()?;
+                rfc4193 = Some(s);
+                records.push(s.to_ptr_soa_name()?);
+            }
+        }
+
         for member in members {
-            let member_name = format!(
-                "zt-{}",
-                member.node_id.expect("Node ID for member does not exist")
-            );
-            let fqdn = member_name.to_fqdn(self.domain_name.clone())?;
-
-            // this is default the zt-<member id> but can switch to a named name if
-            // tweaked in central. see below.
-            let mut canonical_name = fqdn.clone();
-            let mut member_is_named = false;
-
-            if let Some(name) = parse_member_name(member.name.clone(), self.domain_name.clone()) {
-                canonical_name = name;
-                member_is_named = true;
-            }
-
-            let ips: Vec<IpAddr> = member
-                .config
-                .expect("Member config does not exist")
-                .ip_assignments
-                .expect("IP assignments for member do not exist")
-                .into_iter()
-                .map(|s| IpAddr::from_str(&s).expect("Could not parse IP address"))
-                .collect();
-
-            self.match_or_insert(fqdn.clone(), ips.clone());
-
-            if member_is_named {
-                self.match_or_insert(canonical_name.clone(), ips.clone());
-            }
-
-            if let Some(local_ptr_authority) = self.ptr_authority.to_owned() {
-                for ip in ips.clone() {
-                    records.push(ip.into_name().expect("Could not coerce IP into name"));
-                    configure_ptr(local_ptr_authority.clone(), ip, canonical_name.clone())?;
-                }
-            }
-
-            records.push(fqdn.clone());
-            records.push(canonical_name.clone());
+            let record = ZTRecord::new(
+                &member,
+                sixplane,
+                rfc4193,
+                self.domain_name.clone(),
+                self.wildcard_everything,
+            )?;
+            record.insert_authority(self)?;
+            record.insert_member_ptr(
+                self.ptr_authority_map.to_owned(),
+                sixplane,
+                rfc4193,
+                &mut records,
+            )?;
+            record.insert_records(&mut records);
         }
 
         prune_records(self.authority.to_owned(), records.clone())?;
-
-        if let Some(ptr_authority) = self.ptr_authority.to_owned() {
-            prune_records(ptr_authority, records.clone())?;
+        for authority in self.ptr_authority_map.values() {
+            prune_records(authority.to_owned(), records.clone())?;
         }
 
         Ok(())
@@ -489,7 +707,7 @@ impl ZTAuthority {
                         new_rset.add_rdata(rdata);
                     }
 
-                    eprintln!("Replacing host record for {} with {:?}", key, ips);
+                    warn!("Replacing host record for {} with {:?}", key, ips);
                     rr.remove(&rrkey);
                     rr.insert(rrkey, Arc::new(new_rset));
                 }
@@ -498,5 +716,7 @@ impl ZTAuthority {
     }
 }
 
-#[cfg(test)]
+#[cfg(all(feature = "integration-tests", test))]
+mod service;
+#[cfg(all(feature = "integration-tests", test))]
 mod tests;
