@@ -1,25 +1,29 @@
 use std::{
-    collections::HashMap, io::Write, str::FromStr, sync::Arc, thread::sleep, time::Duration,
+    collections::HashMap, io::Write, net::SocketAddr, str::FromStr, sync::Arc, thread::sleep,
+    time::Duration,
 };
 
 use clap::clap_app;
 
 use anyhow::anyhow;
 use ipnetwork::IpNetwork;
+use log::{error, info, warn};
 use tokio::sync::RwLock;
 
 use crate::{
-    authority::{find_members, init_trust_dns_authority, new_ptr_authority},
-    utils::update_central_dns,
+    addresses::Calculator,
+    authority::{find_members, init_trust_dns_authority, new_ptr_authority, ZTAuthority},
+    utils::{central_config, central_token, update_central_dns},
 };
 
+mod addresses;
 mod authority;
 mod hosts;
 mod server;
 mod supervise;
 mod utils;
 
-#[cfg(test)]
+#[cfg(all(feature = "integration-tests", test))]
 mod integration_tests;
 #[cfg(test)]
 mod tests;
@@ -32,59 +36,46 @@ fn write_help(app: clap::App) -> Result<(), anyhow::Error> {
     return Ok(());
 }
 
-fn unsupervise(network: Option<&str>) -> Result<(), anyhow::Error> {
-    supervise::Properties::new(None, network, None, None, None)?.uninstall_supervisor()
+fn unsupervise(args: &clap::ArgMatches<'_>) -> Result<(), anyhow::Error> {
+    supervise::Properties::from(args).uninstall_supervisor()
 }
 
-fn supervise(
-    domain: Option<&str>,
-    network: Option<&str>,
-    hosts_file: Option<&str>,
-    authtoken: Option<&str>,
-    token: Option<&str>,
-) -> Result<(), anyhow::Error> {
-    supervise::Properties::new(domain, network, hosts_file, authtoken, token)?.install_supervisor()
+fn supervise(args: &clap::ArgMatches<'_>) -> Result<(), anyhow::Error> {
+    supervise::Properties::from(args).install_supervisor()
 }
 
-fn start(
-    domain: Option<&str>,
-    network: Option<&str>,
-    hosts_file: Option<&str>,
-    authtoken: Option<&str>,
-    token: Option<&str>,
-) -> Result<(), anyhow::Error> {
+fn start(args: &clap::ArgMatches<'_>) -> Result<(), anyhow::Error> {
+    let domain = args.value_of("domain");
+    let authtoken = args.value_of("secret_file");
+    let network_id = args.value_of("NETWORK_ID");
+    let hosts_file = args.value_of("file");
+    let token = args.value_of("token_file");
+    let wildcard_names = args.is_present("wildcard");
+
     let domain_name = utils::domain_or_default(domain)?;
     let authtoken = utils::authtoken_path(authtoken);
     let runtime = &mut utils::init_runtime();
 
-    if let Some(network) = network {
-        let token = utils::central_token(token);
-        let network = String::from(network);
-        let hf = if let Some(hf) = hosts_file {
+    if let Some(network_id) = network_id {
+        let token = central_config(central_token(token)?);
+        let network_id = String::from(network_id);
+
+        let hosts_file = if let Some(hf) = hosts_file {
             Some(hf.to_string())
         } else {
             None
         };
 
-        if token.is_none() {
-            return Err(anyhow!("missing zerotier central token: set ZEROTIER_CENTRAL_TOKEN in environment, or pass a file containing it with -t"));
-        }
-
-        let token = token.unwrap();
-
-        println!("Welcome to ZeroNS!");
-
-        let ips = runtime.block_on(utils::get_listen_ips(&authtoken, &network))?;
+        info!("Welcome to ZeroNS!");
+        let ips = runtime.block_on(utils::get_listen_ips(&authtoken, &network_id))?;
 
         if ips.len() > 0 {
             update_central_dns(
                 runtime,
                 domain_name.clone(),
-                utils::parse_ip_from_cidr(ips.clone().into_iter()
-                    .find(|i| IpNetwork::from_str(i).expect("Could not parse CIDR").is_ipv4())
-                    .expect("Could not find a valid IPv4 network (currently, ipv6 resolvers are unsupported)")),
+                ips.clone(),
                 token.clone(),
-                network.clone(),
+                network_id.clone(),
             )?;
 
             let mut listen_ips = Vec::new();
@@ -100,37 +91,56 @@ fn start(
                     ipmap.insert(listen_ip, cidr.network());
                 }
 
-                if !authority_map.contains_key(&cidr.network()) {
+                if !authority_map.contains_key(&cidr) {
                     let ptr_authority = new_ptr_authority(cidr)?;
-
-                    let ztauthority = utils::init_authority(
-                        ptr_authority.clone(),
-                        token.clone(),
-                        network.clone(),
-                        domain_name.clone(),
-                        hf.clone(),
-                        Duration::new(30, 0),
-                        authority.clone(),
-                    );
-
-                    let arc_authority = Arc::new(RwLock::new(ztauthority));
-
-                    authority_map.insert(cidr.network(), arc_authority.clone());
-                    runtime.spawn(find_members(arc_authority));
+                    authority_map.insert(cidr, ptr_authority);
                 }
             }
 
-            for ip in listen_ips {
-                println!("Your IP for this network: {}", ip);
-                let cidr = ipmap
-                    .get(&ip)
-                    .expect("Could not locate underlying network subnet");
-                let authority = authority_map
-                    .get(cidr)
-                    .expect("Could not locate PTR authority for subnet");
+            let network = runtime.block_on(
+                zerotier_central_api::apis::network_api::get_network_by_id(&token, &network_id),
+            )?;
 
-                let server = crate::server::Server::new(authority.to_owned());
-                runtime.spawn(server.listen(format!("{}:53", ip.clone()), Duration::new(0, 1000)));
+            let v6assign = network.config.clone().unwrap().v6_assign_mode;
+            if v6assign.is_some() {
+                let v6assign = v6assign.unwrap().clone();
+
+                if v6assign.var_6plane.unwrap_or(false) {
+                    warn!("6PLANE PTR records are not yet supported");
+                }
+
+                if v6assign.rfc4193.unwrap_or(false) {
+                    let cidr = network.clone().rfc4193().unwrap();
+                    if !authority_map.contains_key(&cidr) {
+                        let ptr_authority = new_ptr_authority(cidr)?;
+                        authority_map.insert(cidr, ptr_authority);
+                    }
+                }
+            }
+
+            let mut ztauthority = ZTAuthority::new(
+                domain_name.clone(),
+                network_id.clone(),
+                token.clone(),
+                hosts_file.clone(),
+                authority_map.clone(),
+                Duration::new(30, 0),
+                authority.clone(),
+            );
+
+            if wildcard_names {
+                ztauthority.wildcard_everything();
+            }
+
+            let arc_authority = Arc::new(RwLock::new(ztauthority));
+
+            runtime.spawn(find_members(arc_authority.clone()));
+
+            for ip in listen_ips {
+                info!("Your IP for this network: {}", ip);
+
+                let server = crate::server::Server::new(arc_authority.to_owned());
+                runtime.spawn(server.listen(SocketAddr::new(ip, 53), Duration::new(0, 1000)));
             }
 
             async fn wait() {
@@ -161,6 +171,8 @@ fn main() -> Result<(), anyhow::Error> {
             (@arg file: -f --file +takes_value "An additional lists of hosts in /etc/hosts format")
             (@arg secret_file: -s --secret +takes_value "Path to authtoken.secret (usually detected)")
             (@arg token_file: -t --token +takes_value "Path to a file containing the ZeroTier Central token")
+            (@arg wildcard: -w --wildcard "Wildcard all names in Central to point at the respective member's IP address(es)")
+            (@arg verbose: -v +multiple "Verbose logging (repeat -v for more verbosity")
             (@arg NETWORK_ID: +required "Network ID to query")
         )
         (@subcommand supervise =>
@@ -169,6 +181,7 @@ fn main() -> Result<(), anyhow::Error> {
             (@arg file: -f --file +takes_value "An additional lists of hosts in /etc/hosts format")
             (@arg secret_file: -s --secret +takes_value "Path to authtoken.secret (usually detected)")
             (@arg token_file: -t --token +takes_value +required "Path to a file containing the ZeroTier Central token; this file must not be moved")
+            (@arg wildcard: -w --wildcard "Wildcard all names in Central to point at the respective member's IP address(es)")
             (@arg NETWORK_ID: +required "Network ID to query")
         )
         (@subcommand unsupervise =>
@@ -185,22 +198,17 @@ fn main() -> Result<(), anyhow::Error> {
         None => return write_help(app),
     };
 
-    match cmd {
-        "start" => start(
-            args.value_of("domain"),
-            args.value_of("NETWORK_ID"),
-            args.value_of("file"),
-            args.value_of("secret_file"),
-            args.value_of("token_file"),
-        )?,
-        "supervise" => supervise(
-            args.value_of("domain"),
-            args.value_of("NETWORK_ID"),
-            args.value_of("file"),
-            args.value_of("secret_file"),
-            args.value_of("token_file"),
-        )?,
-        "unsupervise" => unsupervise(args.value_of("NETWORK_ID"))?,
+    stderrlog::new()
+        .module(String::from("zeronsd"))
+        .verbosity((args.occurrences_of("verbose") + 2) as usize)
+        .timestamp(stderrlog::Timestamp::Off)
+        .init()
+        .unwrap();
+
+    let result = match cmd {
+        "start" => start(args),
+        "supervise" => supervise(args),
+        "unsupervise" => unsupervise(args),
         _ => {
             let stderr = std::io::stderr();
             let mut lock = stderr.lock();
@@ -210,6 +218,10 @@ fn main() -> Result<(), anyhow::Error> {
             writeln!(lock)?;
             return Ok(());
         }
+    };
+
+    if result.is_err() {
+        error!("{}", result.unwrap_err())
     }
 
     Ok(())
