@@ -1,3 +1,5 @@
+use std::{net::IpAddr, str::FromStr};
+
 use crate::utils::{domain_or_default, ToHostname};
 
 pub const HOSTS_DIR: &str = "testdata/hosts-files";
@@ -61,7 +63,7 @@ fn test_parse_ip_from_cidr() {
     for (cidr, ip) in results {
         assert_eq!(
             parse_ip_from_cidr(String::from(cidr)),
-            String::from(ip),
+            IpAddr::from_str(ip).unwrap(),
             "{}",
             cidr
         );
@@ -98,14 +100,13 @@ fn test_domain_or_default() {
 fn test_central_token() {
     use crate::utils::central_token;
 
-    assert!(central_token(None).is_none());
+    assert!(central_token(None).is_err());
     std::env::set_var("ZEROTIER_CENTRAL_TOKEN", "abcdef");
-    assert!(central_token(None).is_some());
     assert_eq!(central_token(None).unwrap(), "abcdef");
 
     let hosts = std::fs::read_to_string("/etc/hosts").unwrap();
     let token = central_token(Some("/etc/hosts"));
-    assert!(token.is_some());
+    assert!(token.is_ok());
     assert_eq!(token.unwrap(), hosts.trim());
 }
 
@@ -113,7 +114,7 @@ fn test_central_token() {
 #[should_panic]
 fn test_central_token_panic() {
     use crate::utils::central_token;
-    central_token(Some("/nonexistent"));
+    central_token(Some("/nonexistent")).unwrap();
 }
 
 #[test]
@@ -138,6 +139,7 @@ fn test_supervise_systemd_green() {
                 domain: Some(String::from("zerotier")),
                 authtoken: Some(String::from("/var/lib/zerotier-one/authtoken.secret")),
                 hosts_file: Some(String::from("/etc/hosts")),
+                wildcard_names: true,
             },
         ),
     ];
@@ -160,17 +162,17 @@ fn test_supervise_systemd_green() {
             assert!(path.is_ok(), "{}", name);
             let expected = std::fs::read_to_string(path.unwrap());
             assert!(expected.is_ok(), "{}", name);
-            let testing = props.systemd_template();
+            let testing = props.supervise_template();
             assert!(testing.is_ok(), "{}", name);
 
             assert_eq!(testing.unwrap(), expected.unwrap(), "{}", name);
         } else {
             assert!(props.validate().is_ok(), "{}", name);
 
-            let template = props.systemd_template();
+            let template = props.supervise_template();
             assert!(template.is_ok(), "{}", name);
             assert!(
-                std::fs::write(path, props.systemd_template().unwrap()).is_ok(),
+                std::fs::write(path, props.supervise_template().unwrap()).is_ok(),
                 "{}",
                 name
             );
